# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/etrc',
                'doc_host': 'https://dsm-72.github.io',
                'git_url': 'https://github.com/dsm-72/etrc',
                'lib_path': 'etrc'},
  'syms': { 'etrc.atyp': {},
            'etrc.cons': {},
            'etrc.emod': { 'etrc.emod.ModuleEnum': ('emod.html#moduleenum', 'etrc/emod.py'),
                           'etrc.emod.ModuleEnum._missing_': ('emod.html#moduleenum._missing_', 'etrc/emod.py'),
                           'etrc.emod.ModuleEnum.get': ('emod.html#moduleenum.get', 'etrc/emod.py'),
                           'etrc.emod.ModuleEnum.imp': ('emod.html#moduleenum.imp', 'etrc/emod.py'),
                           'etrc.emod.ModuleEnum.init': ('emod.html#moduleenum.init', 'etrc/emod.py'),
                           'etrc.emod.ModuleEnum.make': ('emod.html#moduleenum.make', 'etrc/emod.py'),
                           'etrc.emod.ModuleEnum.module': ('emod.html#moduleenum.module', 'etrc/emod.py'),
                           'etrc.emod.ModuleEnum.safe': ('emod.html#moduleenum.safe', 'etrc/emod.py')},
            'etrc.enum': { 'etrc.enum.CatOrder': ('enum.html#catorder', 'etrc/enum.py'),
                           'etrc.enum.CatOrder._missing_': ('enum.html#catorder._missing_', 'etrc/enum.py'),
                           'etrc.enum.CatOrder.aidx': ('enum.html#catorder.aidx', 'etrc/enum.py'),
                           'etrc.enum.CatOrder.aug': ('enum.html#catorder.aug', 'etrc/enum.py'),
                           'etrc.enum.CatOrder.augsize': ('enum.html#catorder.augsize', 'etrc/enum.py'),
                           'etrc.enum.CatOrder.augzeros': ('enum.html#catorder.augzeros', 'etrc/enum.py'),
                           'etrc.enum.CatOrder.cat': ('enum.html#catorder.cat', 'etrc/enum.py'),
                           'etrc.enum.CatOrder.iasdim': ('enum.html#catorder.iasdim', 'etrc/enum.py'),
                           'etrc.enum.CatOrder.iindims': ('enum.html#catorder.iindims', 'etrc/enum.py'),
                           'etrc.enum.CatOrder.safe': ('enum.html#catorder.safe', 'etrc/enum.py'),
                           'etrc.enum.Channels': ('enum.html#channels', 'etrc/enum.py'),
                           'etrc.enum.Channels._missing_': ('enum.html#channels._missing_', 'etrc/enum.py'),
                           'etrc.enum.Channels.cidx': ('enum.html#channels.cidx', 'etrc/enum.py'),
                           'etrc.enum.Channels.flip': ('enum.html#channels.flip', 'etrc/enum.py'),
                           'etrc.enum.Channels.resize': ('enum.html#channels.resize', 'etrc/enum.py'),
                           'etrc.enum.Channels.safe': ('enum.html#channels.safe', 'etrc/enum.py'),
                           'etrc.enum.Channels.to': ('enum.html#channels.to', 'etrc/enum.py'),
                           'etrc.enum.DiffusionMethod': ('enum.html#diffusionmethod', 'etrc/enum.py'),
                           'etrc.enum.DiffusionMethod._missing_': ('enum.html#diffusionmethod._missing_', 'etrc/enum.py'),
                           'etrc.enum.DiffusionMethod.imp': ('enum.html#diffusionmethod.imp', 'etrc/enum.py'),
                           'etrc.enum.DynamicsMethod': ('enum.html#dynamicsmethod', 'etrc/enum.py'),
                           'etrc.enum.Framework': ('enum.html#framework', 'etrc/enum.py'),
                           'etrc.enum.Framework._missing_': ('enum.html#framework._missing_', 'etrc/enum.py'),
                           'etrc.enum.Framework.safe': ('enum.html#framework.safe', 'etrc/enum.py'),
                           'etrc.enum.Framework.swap': ('enum.html#framework.swap', 'etrc/enum.py'),
                           'etrc.enum.Geodesic': ('enum.html#geodesic', 'etrc/enum.py'),
                           'etrc.enum.Geodesic.imp': ('enum.html#geodesic.imp', 'etrc/enum.py'),
                           'etrc.enum.InitMethod': ('enum.html#initmethod', 'etrc/enum.py'),
                           'etrc.enum.InitMethod.imp': ('enum.html#initmethod.imp', 'etrc/enum.py'),
                           'etrc.enum.LocReturn': ('enum.html#locreturn', 'etrc/enum.py'),
                           'etrc.enum.NonLinearity': ('enum.html#nonlinearity', 'etrc/enum.py'),
                           'etrc.enum.ODETypes': ('enum.html#odetypes', 'etrc/enum.py'),
                           'etrc.enum.ODETypes._missing_': ('enum.html#odetypes._missing_', 'etrc/enum.py'),
                           'etrc.enum.OTMethod': ('enum.html#otmethod', 'etrc/enum.py'),
                           'etrc.enum.OTMethod._missing_': ('enum.html#otmethod._missing_', 'etrc/enum.py'),
                           'etrc.enum.OTMethod.calc': ('enum.html#otmethod.calc', 'etrc/enum.py'),
                           'etrc.enum.OTMethod.imp': ('enum.html#otmethod.imp', 'etrc/enum.py'),
                           'etrc.enum.OTMethod.loss': ('enum.html#otmethod.loss', 'etrc/enum.py'),
                           'etrc.enum.RecurrentLayer': ('enum.html#recurrentlayer', 'etrc/enum.py'),
                           'etrc.enum.SDENoiseType': ('enum.html#sdenoisetype', 'etrc/enum.py'),
                           'etrc.enum.SDENoiseType._missing_': ('enum.html#sdenoisetype._missing_', 'etrc/enum.py'),
                           'etrc.enum.SDETypes': ('enum.html#sdetypes', 'etrc/enum.py'),
                           'etrc.enum.SDETypes._missing_': ('enum.html#sdetypes._missing_', 'etrc/enum.py'),
                           'etrc.enum.SampleCategoriesMethod': ('enum.html#samplecategoriesmethod', 'etrc/enum.py'),
                           'etrc.enum.SliceFormat': ('enum.html#sliceformat', 'etrc/enum.py'),
                           'etrc.enum.SliceFormat.flip': ('enum.html#sliceformat.flip', 'etrc/enum.py'),
                           'etrc.enum.SliceFormat.guess': ('enum.html#sliceformat.guess', 'etrc/enum.py'),
                           'etrc.enum.SliceFormat.safe': ('enum.html#sliceformat.safe', 'etrc/enum.py'),
                           'etrc.enum.SliceFormat.safeflip': ('enum.html#sliceformat.safeflip', 'etrc/enum.py'),
                           'etrc.enum.SliceFormat.to': ('enum.html#sliceformat.to', 'etrc/enum.py'),
                           'etrc.enum.Solver': ('enum.html#solver', 'etrc/enum.py'),
                           'etrc.enum.Solver._missing_': ('enum.html#solver._missing_', 'etrc/enum.py'),
                           'etrc.enum.Stage': ('enum.html#stage', 'etrc/enum.py'),
                           'etrc.enum.Stage._missing_': ('enum.html#stage._missing_', 'etrc/enum.py'),
                           'etrc.enum.Stage.safe': ('enum.html#stage.safe', 'etrc/enum.py'),
                           'etrc.enum.augidx': ('enum.html#augidx', 'etrc/enum.py'),
                           'etrc.enum.augsize': ('enum.html#augsize', 'etrc/enum.py'),
                           'etrc.enum.augzeros': ('enum.html#augzeros', 'etrc/enum.py'),
                           'etrc.enum.iasdim': ('enum.html#iasdim', 'etrc/enum.py'),
                           'etrc.enum.iindims': ('enum.html#iindims', 'etrc/enum.py')},
            'etrc.kwds': { 'etrc.kwds.diffusion_kwds': ('kwds.html#diffusion_kwds', 'etrc/kwds.py'),
                           'etrc.kwds.recurrent_decoder_kwds': ('kwds.html#recurrent_decoder_kwds', 'etrc/kwds.py'),
                           'etrc.kwds.recurrent_encoder_kwds': ('kwds.html#recurrent_encoder_kwds', 'etrc/kwds.py'),
                           'etrc.kwds.recurrent_kwds': ('kwds.html#recurrent_kwds', 'etrc/kwds.py')},
            'etrc.meta': { 'etrc.meta.ModuleEnumMeta': ('meta.html#moduleenummeta', 'etrc/meta.py'),
                           'etrc.meta.ModuleEnumMeta.__call__': ('meta.html#moduleenummeta.__call__', 'etrc/meta.py'),
                           'etrc.meta.ModuleEnumMeta.__contains__': ('meta.html#moduleenummeta.__contains__', 'etrc/meta.py'),
                           'etrc.meta.ModuleEnumMeta.__copydict__': ('meta.html#moduleenummeta.__copydict__', 'etrc/meta.py'),
                           'etrc.meta.ModuleEnumMeta.__new__': ('meta.html#moduleenummeta.__new__', 'etrc/meta.py'),
                           'etrc.meta.ModuleEnumMeta.__prepare__': ('meta.html#moduleenummeta.__prepare__', 'etrc/meta.py'),
                           'etrc.meta.ModuleEnumMeta._check_for_existing_members_': ( 'meta.html#moduleenummeta._check_for_existing_members_',
                                                                                      'etrc/meta.py'),
                           'etrc.meta.ModuleEnumMeta._default_': ('meta.html#moduleenummeta._default_', 'etrc/meta.py'),
                           'etrc.meta.ModuleEnumMeta._missing_': ('meta.html#moduleenummeta._missing_', 'etrc/meta.py'),
                           'etrc.meta.ModuleEnumMeta._module_': ('meta.html#moduleenummeta._module_', 'etrc/meta.py'),
                           'etrc.meta.ModuleEnumMeta.duplicate': ('meta.html#moduleenummeta.duplicate', 'etrc/meta.py')},
            'etrc.solv': { 'etrc.solv.BaseODE': ('solv.html#baseode', 'etrc/solv.py'),
                           'etrc.solv.BaseSDE': ('solv.html#basesde', 'etrc/solv.py'),
                           'etrc.solv.SolverKind': ('solv.html#solverkind', 'etrc/solv.py'),
                           'etrc.solv.SolverKind._missing_': ('solv.html#solverkind._missing_', 'etrc/solv.py'),
                           'etrc.solv.SolverModule': ('solv.html#solvermodule', 'etrc/solv.py'),
                           'etrc.solv.SolverModule._missing_': ('solv.html#solvermodule._missing_', 'etrc/solv.py'),
                           'etrc.solv.SolverSpecification': ('solv.html#solverspecification', 'etrc/solv.py'),
                           'etrc.solv.SolverSpecification.__contains__': ('solv.html#solverspecification.__contains__', 'etrc/solv.py'),
                           'etrc.solv.SolverSpecification.__eq__': ('solv.html#solverspecification.__eq__', 'etrc/solv.py'),
                           'etrc.solv.SolverSpecification.__init__': ('solv.html#solverspecification.__init__', 'etrc/solv.py'),
                           'etrc.solv.SolverSpecification.__repr__': ('solv.html#solverspecification.__repr__', 'etrc/solv.py'),
                           'etrc.solv.SolverSpecification.__str__': ('solv.html#solverspecification.__str__', 'etrc/solv.py'),
                           'etrc.solv.SolverSpecification.asattr': ('solv.html#solverspecification.asattr', 'etrc/solv.py'),
                           'etrc.solv.SolverSpecification.askey': ('solv.html#solverspecification.askey', 'etrc/solv.py'),
                           'etrc.solv.SolverSpecification.lib': ('solv.html#solverspecification.lib', 'etrc/solv.py'),
                           'etrc.solv.SolverSpecification.load': ('solv.html#solverspecification.load', 'etrc/solv.py'),
                           'etrc.solv.SolverSpecification.match': ('solv.html#solverspecification.match', 'etrc/solv.py'),
                           'etrc.solv.SolverSpecification.names': ('solv.html#solverspecification.names', 'etrc/solv.py'),
                           'etrc.solv.SolverSpecification.regattr': ('solv.html#solverspecification.regattr', 'etrc/solv.py'),
                           'etrc.solv.SolverSpecification.regkey': ('solv.html#solverspecification.regkey', 'etrc/solv.py'),
                           'etrc.solv.SolverSystem': ('solv.html#solversystem', 'etrc/solv.py'),
                           'etrc.solv.SolverSystem._missing_': ('solv.html#solversystem._missing_', 'etrc/solv.py'),
                           'etrc.solv.Solvers': ('solv.html#solvers', 'etrc/solv.py'),
                           'etrc.solv.Solvers.__contains__': ('solv.html#solvers.__contains__', 'etrc/solv.py'),
                           'etrc.solv.Solvers.__str__': ('solv.html#solvers.__str__', 'etrc/solv.py'),
                           'etrc.solv.Solvers.all': ('solv.html#solvers.all', 'etrc/solv.py'),
                           'etrc.solv.Solvers.asattr': ('solv.html#solvers.asattr', 'etrc/solv.py'),
                           'etrc.solv.Solvers.askey': ('solv.html#solvers.askey', 'etrc/solv.py'),
                           'etrc.solv.Solvers.attrs': ('solv.html#solvers.attrs', 'etrc/solv.py'),
                           'etrc.solv.Solvers.get': ('solv.html#solvers.get', 'etrc/solv.py'),
                           'etrc.solv.Solvers.items': ('solv.html#solvers.items', 'etrc/solv.py'),
                           'etrc.solv.Solvers.keys': ('solv.html#solvers.keys', 'etrc/solv.py'),
                           'etrc.solv.Solvers.lienent_equality': ('solv.html#solvers.lienent_equality', 'etrc/solv.py'),
                           'etrc.solv.Solvers.matchdict': ('solv.html#solvers.matchdict', 'etrc/solv.py'),
                           'etrc.solv.Solvers.matchtuple': ('solv.html#solvers.matchtuple', 'etrc/solv.py'),
                           'etrc.solv.Solvers.odes': ('solv.html#solvers.odes', 'etrc/solv.py'),
                           'etrc.solv.Solvers.register': ('solv.html#solvers.register', 'etrc/solv.py'),
                           'etrc.solv.Solvers.sdes': ('solv.html#solvers.sdes', 'etrc/solv.py'),
                           'etrc.solv.Solvers.values': ('solv.html#solvers.values', 'etrc/solv.py')},
            'etrc.util': { 'etrc.util.enumdict': ('util.html#enumdict', 'etrc/util.py'),
                           'etrc.util.evalue': ('util.html#evalue', 'etrc/util.py'),
                           'etrc.util.get_torch_module': ('util.html#get_torch_module', 'etrc/util.py'),
                           'etrc.util.missing': ('util.html#missing', 'etrc/util.py'),
                           'etrc.util.nextv': ('util.html#nextv', 'etrc/util.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/05_solv.ipynb.

# %% auto 0
__all__ = ['EulerODE', 'EulerSDE', 'SRK', 'MidpointODE', 'MidpointSDE', 'LogODEMidpoint', 'Heun', 'EulerHeun', 'ReversibleHeun',
           'AdjointReversibleHeun', 'MilsteinIto', 'MilsteinStratonovich', 'RungeKutta4', 'DormandPrince45',
           'Tsitouras45', 'ImplicitEuler', 'AsynchronousLeapfrog', 'SolverSystem', 'SolverModule', 'SolverKind',
           'Solvers', 'SolverSpecification', 'BaseODE', 'BaseSDE']

# %% ../nbs/05_solv.ipynb 6
from importlib import import_module
from enum import StrEnum, auto
from itertools import zip_longest

# %% ../nbs/05_solv.ipynb 8
#| export


# %% ../nbs/05_solv.ipynb 10
#| export


# %% ../nbs/05_solv.ipynb 12
from nlit import NAME, SYSTEM, MODULE, KIND

# %% ../nbs/05_solv.ipynb 14
from .util import missing

# %% ../nbs/05_solv.ipynb 17
class SolverSystem(StrEnum):
    '''Enumeration for solver systems to categorize between ordinary differential equations (ODE) 
    and stochastic differential equations (SDE).
    
    Attributes
    ----------
    ode : str
        Represents the ordinary differential equation system.
    
    sde : str 
        Represents the stochastic differential equation system.
        
    Methods
    -------
    _missing_(val: str):
        A class method to handle missing values by returning a default value.
    
    Examples
    --------
    >>> SolverSystem.ode
    SolverSystem.ode
    >>> SolverSystem('sde')
    SolverSystem.sde
    '''
    
    ode = auto();
    '''Represents the ordinary differential equation system.'''
    
    sde = auto();
    '''Represents the stochastic differential equation system.'''
    
    @classmethod
    def _missing_(cls, val: str): return missing(cls, val, cls.ode)


# %% ../nbs/05_solv.ipynb 19
class SolverModule(StrEnum):
    '''Enumeration for solver modules, mapping each solver system to its 
    corresponding implementation module.
    
    Attributes
    ----------
    ode : str
        Module path for ODE solvers.
    sde : str
        Module path for SDE solvers.
        
    Examples
    --------
    >>> SolverModule.ode
    'torchdyn.numerics.solvers.ode'
    >>> SolverModule.sde
    'torchsde._core.methods'
    '''
    
    ode = 'torchdyn.numerics.solvers.ode'; 
    '''Module path for ODE solvers (i.e. `'torchdyn.numerics.solvers.ode'`).'''
    
    sde = 'torchsde._core.methods';
    '''Module path for SDE solvers (i.e. `'torchsde._core.methods'`).'''
    
    @classmethod
    def _missing_(cls, val: str): return missing(cls, val, cls.ode)
    

# %% ../nbs/05_solv.ipynb 21
class SolverKind(StrEnum):
    '''Enumeration for solver kinds, differentiating between Ito and Stratonovich interpretations for SDEs.
    
    Attributes
    ----------
    ito : str
        Represents the Ito interpretation.
    
    stratonovich : str
        Represents the Stratonovich interpretation.
        
    Examples
    --------
    >>> SolverKind.ito
    SolverKind.ito
    >>> SolverKind.stratonovich
    SolverKind.stratonovich
    '''
    
    ito = auto(); 
    '''Represents the Ito interpretation.'''
    
    stratonovich = auto();
    '''Represents the Stratonovich interpretation.'''
    
    @classmethod
    def _missing_(cls, val: str): return missing(cls, val, cls.ito)

# %% ../nbs/05_solv.ipynb 23
class Solvers:
    '''Central registry for all solver specifications.
    
    This class allows for dynamic registration and querying of solver specifications, supporting advanced lookup and management functionalities.
    
    Methods
    -------
    keys() -> List[str] 
        Returns a sorted list of all registered solver keys.
        
    values() -> List[SolverSpecification]
        Returns a list of all solver instances in the registry.
        
    attrs() -> List[str]
        Returns a list of attribute names for all solvers.
        
    items() -> Dict[str, SolverSpecification]
        Returns a dictionary mapping attribute names to solver instances.
    
    all() -> List[str]
        Returns a list of string representations for all registered solvers.
    
    odes() -> List[str]
        Filters and returns solvers for ODE systems.
    
    sdes() -> List[str]
        Filters and returns solvers for SDE systems.
    
    get(key) -> SolverSpecification
        Retrieves a solver instance by key, supporting flexible key types.
    
    askey(solv)
        Returns the registry key for a given solver specification.
        
    asattr(solv)
        Returns the attribute name for a given solver specification.
        
    matchdict(key)
        Matches a dictionary to the closest registered solver specification key.
        
    lienent_equality(a, b)
        Checks for lenient equality between two solver keys.
        
    matchtuple(key)
        Matches a tuple to the closest registered solver specification key.
    
    register(solv: SolverSpecification) -> SolverSpecification
        Registers a new solver specification or updates an existing one.
        
    Examples
    --------
    >>> Solvers.register(BaseODE(name='CustomODESolver'))
    >>> 'CustomODESolver' in Solvers
    True
    '''
    _registry = {}
    
    @classmethod
    def keys(cls) -> list:
        '''Returns a sorted list of all solver keys registered.

        Returns
        -------
        list
            Sorted list of solver keys.
        '''
        return sorted([v.regkey for v in cls._registry.values()])
    
    @classmethod
    def values(cls) -> list: 
        '''Returns all solver instances in the registry.

        Returns
        -------
        list
            List of solver instances.
        '''
        return [cls._registry[k] for k in cls.keys()]
    
    @classmethod
    def attrs(cls) -> list:
        '''Returns the attribute names for all registered solver instances.

        Returns
        -------
        list
            List of attribute names.
        '''
        return [v.regattr for v in cls.values()]
    
    @classmethod
    def items(cls) -> dict: 
        '''Returns a dictionary mapping attribute names to solver instances.

        Returns
        -------
        dict
            Dictionary of attribute names to solver instances.
        '''
        return {cls._registry[k].regattr: cls._registry[k] for k in cls.keys()}
    
    @classmethod
    def all(cls) -> list[str]:
        '''Returns a list of string representations of all registered solvers.

        Returns
        -------
        list
            List of string representations of solvers.
        '''
        return [str(v) for v in cls.values()]
    
    @classmethod
    def odes(cls) -> list[str]:
        '''Returns a list of ODE solvers.

        Returns
        -------
        list
            List of ODE solvers as strings.
        '''
        return [str(s) for s in cls.values() if s.system == SolverSystem.ode]
    
    @classmethod
    def sdes(cls) -> list[str]: 
        '''Returns a list of SDE solvers.

        Returns
        -------
        list
            List of SDE solvers as strings.
        '''
        return [str(s) for s in cls.values() if s.system == SolverSystem.sde]
    
    def __str__(cls) -> str:
        '''Returns a string representation of all registered solvers.

        Returns
        -------
        str
            String representation of all solvers.
        '''
        return str(cls.all())

    def __contains__(cls, item) -> bool:
        '''Checks if the registry contains a given solver or attribute.

        Parameters
        ----------
        item : str or SolverSpecification
            The item to check for existence in the registry.

        Returns
        -------
        bool
            True if item is found, False otherwise.
        '''
        for sol in cls._registry.values():
            if sol.match(item): return True
        return item in cls.all()
    
    @classmethod
    def get(cls, key: str):
        '''Retrieves a solver instance by key.

        Parameters
        ----------
        key : str, tuple, or dict
            The key to identify the solver. Can be a string (name), tuple (specification), 
            or dict (specification attributes).

        Returns
        -------
        SolverSpecification
            The matching solver instance.

        Raises
        ------
        KeyError
            If no solver matches the key.
        '''
        if isinstance(key, str):
            key = key.casefold()
            for reg, ins in cls._registry.items():
                if ins.match(key) or key == ins.regattr:
                    return ins
            raise KeyError(f'''Solver '{key}' not found.''')
        elif isinstance(key, tuple):
            nkey = cls.matchtuple(key)
            return cls._registry[nkey]
        elif isinstance(key, dict):
            nkey = cls.matchtuple(key)
            return cls._registry[nkey]
        
        else:
            raise KeyError(f'Invalid key type: {type(key)}')
        
    @classmethod
    def askey(cls, solv: 'SolverSpecification'): 
        '''Returns the registry key for a given solver specification.'''
        return solv.regkey
    
    @classmethod
    def asattr(cls, solv: 'SolverSpecification'): 
        '''Returns the attribute name for a given solver specification.'''
        return solv.regattr
    
    @classmethod
    def matchdict(cls, key: dict):
        '''Matches a dictionary to the closest registered solver specification key.'''
        defs = (None, SolverSystem.ode, SolverModule.ode, None)
        strs = (key.get(s, None) for s in (NAME, SYSTEM, MODULE, KIND))
        skey = tuple(next(iter(e)) for e in zip_longest(strs, defs))
        return cls.matchtuple(skey)
    
    @classmethod
    def lienent_equality(cls, a: tuple, b: tuple) -> bool:
        '''Checks for lenient equality between two solver keys.'''
        iseq = True
        for (x, y) in zip(a, b):
            if x is None: continue
            if x == y: continue
            if isinstance(x, str) and isinstance(y, str):
                if x.casefold() == y.casefold(): 
                    continue
            iseq = False
            break
        return iseq
    
    @classmethod
    def matchtuple(cls, key: tuple):
        '''Matches a tuple to the closest registered solver specification key.'''
        defs = (None, SolverSystem.ode, SolverModule.ode, None)
        skey = list(next(iter(e)) for e in zip_longest(tuple(key), defs))
        if skey[1] is not None: skey[1] = SolverSystem(skey[1])
        if skey[2] is not None: skey[2] = SolverModule(skey[2])
        skey = tuple(skey)
        for rkey in cls._registry:
            if cls.lienent_equality(skey, rkey):  return rkey
        return skey
    
    @classmethod
    def register(cls, solv: 'SolverSpecification'):
        '''Registers a new solver specification or updates an existing one.'''
        # Check for existing entry with case-insensitive name/synonyms
        skey = solv.regkey
        rkey = cls.matchtuple(skey)
        for rkey, rval in cls._registry.items():
            nam = solv.match(rkey[0])
            sys = solv.system == rval.system 
            mod = solv.module == rval.module 
            knd = solv.kind == rval.kind
            if nam and sys and mod and knd:
                solv = rval
                return solv
        
        # Register the instance
        cls._registry[skey] = solv
        setattr(cls, solv.regattr, solv)
        return solv

# %% ../nbs/05_solv.ipynb 25
class SolverSpecification:
    '''A base class for defining solver specifications, encapsulating solver details.

    Parameters
    ----------
    name : str, optional
        The name of the solver, by default inferred from the class name.
        
    synonyms : set of str, optional
        Alternative names or synonyms for the solver, by default an empty set.
        
    kind : SolverKind, optional
        The kind of solver, applicable for SDEs (Ito or Stratonovich), by default None.
        
    system : SolverSystem, optional
        The system category (ODE or SDE) for the solver, by default derived from the class.
        
    module : SolverModule, optional
        The module where the solver's implementation is located, by default derived from the class.

    Attributes
    ----------
    system : SolverSystem
        The system category of the solver.
        
    module : SolverModule
        The module path for the solver's implementation.
        
    name : str
        The name of the solver.
        
    synonyms : set of str
        A set containing the solver's name, synonyms, and a registry attribute.

    lib : str
        Returns the library part of the module path.
        
    regkey : tuple[str, SolverSystem, SolverModule, SolverKind | None]
        A tuple representing the solver's key for registration.
    
    regattr : str
        A string representing the solver's attribute for registration.
        
    Methods
    -------
    askey(solv)
        Constructs a key tuple from a SolverSpecification instance.
        
    asattr(solv)
        Constructs a registry attribute from a SolverSpecification.
        
    load()
        Dynamically imports and returns the solver class from its module.
        
    names()
        Returns a case-folded set of all names associated with the solver.
        
    match(s)
        Checks if a given string matches any of the solver's names.
        
    __contains__(other)
        Checks if the solver matches or contains another solver specification or name.

    Examples
    --------
    >>> class MyODESolver(SolverSpecification):
    ...     pass
    >>> solver = MyODESolver(name='CustomODESolver')
    >>> print(solver.lib)
    'torchdyn'
    '''
    system: SolverSystem = SolverSystem.ode
    '''The system category of the solver.'''
    
    module: SolverModule = SolverModule.ode
    '''The module path for the solver's implementation.'''
    
    def __init__(
        self, 
        name: str | None = None,         
        synonyms: set[str] = set(), 
        kind: SolverKind | None = None,
        system: SolverSystem | None = None,
        module: SolverModule | None = None,
        *args, **kwargs,
    ):
        cstr = type(self).__name__        
        self.name = name or cstr
        self.synonyms = synonyms | {name, cstr}
        self.system = system or SolverSystem(type(self).system)
        self.module = module or SolverModule(type(self).module)
        self.kind = kind if kind is None else SolverKind(kind)
        self.synonyms |= {self.regattr}
        # Register the instance
        self = Solvers.register(self)
        
    @property
    def lib(self) -> str: 
        return self.module.partition('.')[0]
    
    @classmethod
    def askey(cls, solv: 'SolverSpecification') -> tuple[str, SolverSystem, SolverModule, SolverKind | None]: 
        return (solv.name, solv.system, solv.module, solv.kind)

    @classmethod
    def asattr(cls, solv: 'SolverSpecification') -> str:
        n, s, m, k = cls.askey(solv)
        attr = f'{n}'
        if k is not None: attr += f'{k.capitalize()}'
        if s is not None: attr += f'{s.upper()}'
        return attr
        
    @property
    def regkey(self) -> tuple[str, SolverSystem, SolverModule, SolverKind | None]: 
        return type(self).askey(self)
    
    @property
    def regattr(self) -> str:
        return type(self).asattr(self)
    
    def __str__(self) -> str:
        return self.name
     
    def __repr__(self) -> str:
        return f'{self.name}({self.system.upper()}Solver)'
        
    def load(self) -> type:
        '''Dynamically imports and returns the solver class from its module.'''
        mod = import_module(str(self.module))
        attr = f'{self.name}'
        if self.kind is not None:
            attr = f'{attr}{self.kind.capitalize()}'
        return getattr(mod, attr)

    def names(self) -> set[str]:
        '''Returns a case-folded set of all names associated with the solver.'''
        return {n.casefold() for n in self.synonyms}

    def match(self, s: str) -> bool:
        '''Checks if a given string matches any of the solver's names.'''
        return str(s).casefold() in self.names()

    def __contains__(self, other: str) -> bool:
        '''Checks if the solver matches or contains another solver specification or name.'''
        if isinstance(other, SolverSpecification):
            return any(map(self.match, other.names()))
        return self.match(other)

    def __eq__(self, other):
        match = other in self
        if not isinstance(other, SolverSpecification): return match
        attrs = self.system == other.system and self.kind == other.kind and self.module == other.module
        return match and attrs and self.name == other.name
    

# %% ../nbs/05_solv.ipynb 27
class BaseODE(SolverSpecification):
    '''
    Base class for Ordinary Differential Equation (ODE) solver specifications.

    Inherits all parameters, attributes, and methods from SolverSpecification, 
    setting the default system to ODE and specifying the module for ODE solver implementations.
    '''
    module = SolverModule.ode; system = SolverSystem.ode;
    
class BaseSDE(SolverSpecification):
    '''
    Base class for Stochastic Differential Equation (SDE) solver specifications.

    Inherits all parameters, attributes, and methods from SolverSpecification, 
    setting the default system to SDE and specifying the module for SDE solver implementations.
    '''
    module = SolverModule.sde; system = SolverSystem.sde;

# %% ../nbs/05_solv.ipynb 29
EulerODE = BaseODE('Euler')
EulerSDE = BaseSDE('Euler')
SRK = BaseSDE('SRK')
MidpointODE = BaseODE('Midpoint')
MidpointSDE = BaseSDE('Midpoint')
LogODEMidpoint = BaseSDE('LogODEMidpoint')
Heun = BaseSDE('Heun')
EulerHeun = BaseSDE('EulerHeun')
ReversibleHeun = BaseSDE('ReversibleHeun', synonyms={'rheun'})
AdjointReversibleHeun = BaseSDE('AdjointReversibleHeun', synonyms={'arheun'})
MilsteinIto = BaseSDE('MilsteinIto', kind=SolverKind.ito, synonyms={'milito'})
MilsteinStratonovich = BaseSDE('MilsteinStratonovich', kind=SolverKind.stratonovich, synonyms={'milvich'})
RungeKutta4 = BaseODE('RungeKutta4', synonyms={'rk4', 'rk-4'})
DormandPrince45 = BaseODE('DormandPrince45', synonyms={'dopri5', 'dp45', 'dp-45'})
Tsitouras45 = BaseODE('Tsitouras45', synonyms={'tsit5', 'tsit-5'})
ImplicitEuler = BaseODE('ImplicitEuler', synonyms={'ieuler', 'ie'})
AsynchronousLeapfrog =  BaseODE('AsynchronousLeapfrog', synonyms={'alf'})

# %% ../nbs/05_solv.ipynb 35
#| export



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/10_enum.ipynb.

# %% auto 0
__all__ = ['DimChars', 'DimType', 'DimName', 'dimchars', 'size2str', 'OTMethod', 'SDENoiseType', 'SDETypes', 'DynamicsMethod',
           'ODETypes', 'augzeros', 'CatOrder', 'Solver', 'Stage', 'Channels', 'NonLinearity', 'InitMethod',
           'RecurrentLayer', 'SliceFormat', 'DiffusionMethod', 'Geodesic', 'Framework', 'LocReturn',
           'SampleCategoriesMethod']

# %% ../nbs/10_enum.ipynb 6
from enum import Enum, IntEnum, StrEnum, auto
from functools import partial, wraps
from importlib import import_module

# %% ../nbs/10_enum.ipynb 8
from types import  ModuleType, FunctionType
from typing import (Type, Self, Callable, Optional)

# %% ../nbs/10_enum.ipynb 10
try: import numpy as np
except ImportError: ...

try: import torchvision.transforms.functional as F
except ImportError: ...

try: import torch
except ImportError: ...

try: import ot
except: ...

# %% ../nbs/10_enum.ipynb 12
from nchr import U1, DOT
from nlit import NUMPY, DEVICE, SHAPE, UNBALANCED

from quac import tensor, imagesizeq, slicespecs, deviceq
from quac.eggs import tv_functional, pt

from chck import isnone, notnone, isslice, isarray, istensor, isint

from asto import asdev, asten, asarr, to
from dstr import (
    DimChars as _DimChars, DimType as _DimType, DimName as _DimName, 
    dimchars as _dimchars, size2str as _size2str
)

# %% ../nbs/10_enum.ipynb 14
from .cons import _SUFFIX
from .atyp import T, P
from .util import missing
from .emod import ModuleEnum

from etrc.solv import (
    EulerODE, EulerSDE, SRK, MidpointODE, MidpointSDE, LogODEMidpoint, Heun, EulerHeun, 
    ReversibleHeun, AdjointReversibleHeun, MilsteinIto, MilsteinStratonovich, RungeKutta4, 
    DormandPrince45, Tsitouras45, ImplicitEuler, AsynchronousLeapfrog
)

# %% ../nbs/10_enum.ipynb 16
class OTMethod(ModuleEnum, module = 'ot', default = 'emd'):
    '''Enumeration for specifying OT Method.

    Inherits from ModuleEnum to enable module-specific functionality.
    
    Attributes
    ----------
    emd : auto
        Earth Mover's Distance
        
    sinkhorn: auto
        Sinkhorn algorithm
    
    sinkhorn_knopp_unbalanced: auto
        Unbalanced Sinkhorn algorithm
        
    Methods
    -------
    _missing_(val: str)
        Returns a default solver for missing entries.
     '''
     
    emd = auto()
    sinkhorn = auto()
    sinkhorn_knopp_unbalanced = auto()
    
    @classmethod
    def _missing_(cls, val: str):  
        return missing(cls, val, cls.emd)
    
    @classmethod
    def imp(cls: Type[T], key: T | None = None, *args: P.args, **kwargs: P.kwargs) -> Callable:
        '''Import the class or function corresponding to the enum member, 
        or the default member of the enum class'''
        mth = cls.safe(key)
        if UNBALANCED in str(mth):
            try: mod =import_module(DOT.join((cls._module_(), UNBALANCED)))
            except: mod = None
        else: mod = cls.module()
        return getattr(mod, mth)
    
    def calc(
        self,
        source: tensor,
        target: tensor,
        __fn: FunctionType | None = None,
        *args: P.args, 
        **kwargs: P.kwargs
    ) -> tensor:
        '''
        Calculate the loss via the corresponding enum member's Optimal Transport
        method.'''
        try: import ot
        except: ...
        fn = self.get() if isnone(__fn) else __fn
        
        mu = torch.from_numpy(ot.unif(source.size(0)))
        nu = torch.from_numpy(ot.unif(target.size(0)))
        M = torch.cdist(source, target) ** 2

        pi = fn(mu.cpu(), nu.cpu(), M.detach().cpu())
        
        if istensor(pi): 
            pi = pi.clone().detach()
            
        if isarray(pi): 
            pi = torch.tensor(pi)
            
        pim = to(pi, other=source) * to(M, other=source)
        loss = torch.sum(pim)
        return loss
    
    @classmethod
    def loss(
        cls,
        source: tensor,
        target: tensor,
        method: T | None = None, 
        func: FunctionType | None = None,
        *args: P.args, 
        **kwargs: P.kwargs
    ) -> tensor:
        '''
        Call the Optimal Transport Method corresponding to the enum member, 
        or the default member of the enum class with the given arguments.'''
        inst = cls.safe(method)
        return inst.calc(source, target, *args, __fn = func, **kwargs)

# %% ../nbs/10_enum.ipynb 19
class SDENoiseType(StrEnum):
    '''Enumeration for specifying the noise type in a Stochastic Differential Equation (SDE).

    Attributes
    ----------
    general : auto
        General noise structure.
        
    diagonal : auto
        Noise is diagonal.
        
    scalar : auto
        Noise is scalar.
        
    additive : auto
        Noise is additive.

    Methods
    -------
    _missing_(val: str)
        Provides a default value for missing entries.

    Examples
    --------
    >>> SDENoiseType('general')
    SDENoiseType.general
    >>> SDENoiseType('nonexistent')  # This would trigger the _missing_ method.
    SDENoiseType.diagonal
    '''
    general = auto()
    diagonal = auto()
    scalar = auto()
    additive = auto()
    
    @classmethod
    def _missing_(cls, val: str):
        '''Return a default value for missing entries.'''
        return missing(cls, val, cls.diagonal)

# %% ../nbs/10_enum.ipynb 21
class SDETypes(ModuleEnum, module = 'torchsde._core.methods', default='Euler'):
    '''Enumeration for specifying SDE solver types with predefined solver classes.

    Inherits from ModuleEnum to enable module-specific functionality.

    Attributes are predefined solver classes such as `Euler`, `Heun`, etc.

    Methods
    -------
    _missing_(val: str)
        Returns a default solver for missing entries.

    Examples
    --------
    >>> SDETypes('Euler')
    SDETypes.Euler
    >>> SDETypes('nonexistent')  # This would trigger the _missing_ method.
    SDETypes.Euler
    '''
    AdjointReversibleHeun = AdjointReversibleHeun
    Euler = EulerSDE
    EulerHeun = EulerHeun
    Heun = Heun
    LogODEMidpoint = LogODEMidpoint
    Midpoint = MidpointSDE
    MilsteinIto = MilsteinIto
    MilsteinStratonovich = MilsteinStratonovich
    ReversibleHeun = ReversibleHeun
    SRK = SRK
    @classmethod
    def _missing_(cls, val: str): 
        return missing(cls, val, cls.Euler)

# %% ../nbs/10_enum.ipynb 24
class DynamicsMethod(str, ModuleEnum, module = 'torchdyn.core', default='NeuralODE'):
    '''Enumeration for specifying dynamics methods within the torchdyn library.

    Attributes
    ----------
    NeuralODE : auto
        Represents a neural ordinary differential equation.
    NeuralSDE : auto
        Represents a neural stochastic differential equation.

    Methods
    -------
    None specific beyond inherited.

    Examples
    --------
    >>> DynamicsMethod('NeuralODE')
    DynamicsMethod.NeuralODE
    >>> DynamicsMethod('NeuralSDE')
    DynamicsMethod.NeuralSDE
    '''
    NeuralODE = auto()
    NeuralSDE = auto()

# %% ../nbs/10_enum.ipynb 26
class ODETypes(ModuleEnum, module = 'torchdyn.numerics.solvers.ode', default='Euler'):
    '''Enumeration for specifying ODE solver types with predefined solver classes.

    Inherits from ModuleEnum to enable module-specific functionality.

    Attributes are predefined solver classes like `Euler`, `Midpoint`, etc.

    Methods
    -------
    _missing_(val: str)
        Returns a default solver for missing entries.

    Examples
    --------
    >>> ODETypes('Euler')
    ODETypes.Euler
    >>> ODETypes('nonexistent')  # This would trigger the _missing_ method.
    ODETypes.Euler
    '''
    AsynchronousLeapfrog = AsynchronousLeapfrog
    DormandPrince45 = DormandPrince45
    Euler = EulerODE
    ImplicitEuler = ImplicitEuler
    Midpoint = MidpointODE
    RungeKutta4 = RungeKutta4
    Tsitouras45 = Tsitouras45
    @classmethod
    def _missing_(cls, val: str):  
        return missing(cls, val, cls.Euler)

# %% ../nbs/10_enum.ipynb 28
def iindims(i: int, dims: tuple) -> int:
    '''Ensure that the index `i` is within the bounds of the `dims`.'''
    return abs(i) < len(dims)

def iasdim(i: int, dims: tuple) -> int:
    '''Ensure that the index `i` is within the bounds of the `dims` or return -1.'''
    return i if iindims(i, dims) else -1

def augidx(i: int, dims: int | tuple[int, ...]) -> int:
    '''Safely determine the augment index.'''
    ndim = dims if isint(dims) else len(dims) 
    aidx = aidx if abs(aidx) < ndim else -1
    return aidx

def augsize(x: tensor, idx: int = 1, dim: int = 1, add: bool = False) -> tuple:
    '''Return the size of the tensor `x` if augmented at index `idx` to `dim`.'''
    dims = list(getattr(x, SHAPE, ()))
    aidx = augidx(idx, dims)
    dims =  dims + [None]
    
    # NOTE: dim will never overwrite a dimension if add is True
    if add: dims = dims[:aidx] + [dim,] + dims[aidx:]
        
    # NOTE: dim might overwrite a dimension if add is False
    else: dims[aidx] = dim
    
    if dims[-1] is None: dims.pop()
        
    return tuple(dims)

# %% ../nbs/10_enum.ipynb 29
def augzeros(x: tensor, idx: int = 1, dim: int = 1, add: bool = False) -> tensor:
    try: import torch
    except: ...
    return torch.zeros(augsize(x, idx, dim, add))

# %% ../nbs/10_enum.ipynb 30
class CatOrder(IntEnum):
    '''Integer enumeration to specify concatenation order.

    Attributes
    ----------
    first : 1
        Indicates first order.
    last : -1
        Indicates last order.

    Methods
    -------
    _missing_(n: int)
        Provides a default value for missing or invalid orders.

    Examples
    --------
    >>> CatOrder(1)
    CatOrder.first
    >>> CatOrder(-1)
    CatOrder.last
    >>> CatOrder('first')
    CatOrder.first
    '''
    first =  1  
    last  = -1
    
    @classmethod
    def _missing_(cls, n: int) -> 'CatOrder':
        if str(n) in {'first', '1'}: n = 1
        if str(n) in {'last', '-1'}: n = -1
        return cls.first if isinstance(n, int) and n > 0 else cls.last
    
    @classmethod
    def safe(cls, stage: 'CatOrder') -> 'CatOrder':
        '''Take a key and return the corresponding enum member if it exists, 
        otherwise return the default member'''
        try: return cls(stage)
        except ValueError: return cls.last
        
    @staticmethod
    @wraps(augidx)
    def aidx(idx, dims: int | tuple[int, ...]) -> int:
        return augidx(idx, dims)
    
    @staticmethod
    @wraps(augsize)
    def augsize(x: tensor, idx: int = 1, dim: int = 1, add: bool = False) -> tuple:
        return augsize(x, idx, dim, add)
    
    @staticmethod
    @wraps(iindims)
    def iindims(i: int, dims: tuple) -> int:
        return iindims(i, dims)
    
    @staticmethod
    @wraps(iasdim)
    def iasdim(i: int, dims: tuple) -> int:
        return iasdim(i, dims)
    
    @staticmethod
    @wraps(augzeros)
    def augzeros(x: tensor, idx: int = 1, dim: int = 1, add: bool = False) -> tuple:
        return augzeros(x, idx, dim, add)
    
    
    @classmethod
    def cat(
        cls: Type[Self], 
        x: tensor, 
        y: tensor, 
        idx: int = 1,
        order: Type[Self] = 'last', 
    ) -> tensor | NotImplementedError:
        '''Concatenate two tensors along a specified dimension.'''
        # trc: ModuleType = imod('torch')
        trc: ModuleType = pt.duck
        ord = cls.safe(order)
        match CatOrder(ord):
            case CatOrder.first: return trc.cat([y.to(x), x], idx)
            case CatOrder.last:  return trc.cat([x, y.to(x)], idx)
            case _: raise NotImplementedError(f'CatOrder {order} not implemented.')
    
    @classmethod
    def aug(
        cls: Type[Self], 
        x: tensor, 
        fnc: Optional[Callable[[tensor], tensor]] = None, 
        dim: int = 1,
        idx: int = 1,
        order: Type[Self] = 'last', 
    ) -> tensor | NotImplementedError:
        '''Augment a tensor at index `idx` by `dim` dimensions
        with a augment function `fnc` with the concatenated in `order`.'''
        ord = cls.safe(order)
        idx = cls.aidx(idx)
        fnc = fnc or partial(augzeros, idx=idx, dim=dim)
        return cls.cat(x, fnc(x), idx, ord)
    
    

# %% ../nbs/10_enum.ipynb 33
class Solver(Enum):
    '''Enumeration for solver types, including both ODE and SDE solvers.

    Methods
    -------
    _missing_(val: str)
        Returns a default solver for missing entries.

    Examples
    --------
    >>> Solver('EulerODE')
    Solver.EulerODE
    >>> Solver('nonexistent')  # This would trigger the _missing_ method.
    Solver.EulerODE
    '''
    EulerODE = EulerODE
    EulerSDE = EulerSDE
    SRK = SRK
    MidpointODE = MidpointODE
    MidpointSDE = MidpointSDE
    LogODEMidpoint = LogODEMidpoint
    Heun = Heun
    EulerHeun = EulerHeun
    ReversibleHeun = ReversibleHeun
    AdjointReversibleHeun = AdjointReversibleHeun
    MilsteinIto = MilsteinIto
    MilsteinStratonovich = MilsteinStratonovich
    RungeKutta4 = RungeKutta4
    DormandPrince45 = DormandPrince45
    Tsitouras45 = Tsitouras45
    ImplicitEuler = ImplicitEuler
    AsynchronousLeapfrog = AsynchronousLeapfrog
    @classmethod
    def _missing_(cls, val: str):  
        return missing(cls, val, cls.EulerODE)

# %% ../nbs/10_enum.ipynb 36
class Stage(StrEnum):
    '''Enumeration for specifying the stage in a processing or evaluation pipeline.

    Attributes
    ----------
    fit : auto
        Indicates the fitting or training stage.
        
    validate : auto
        Indicates the validation stage.
        
    test : auto
        Indicates the testing stage.
        
    predict : auto
        Indicates the prediction stage.

    Methods
    -------
    _missing_(val: str)
        Provides a default stage for missing entries.
        
    safe(stage: 'Stage') -> 'Stage'
        Returns the corresponding enum member if it exists, otherwise returns the default stage.

    Examples
    --------
    >>> Stage('fit')
    Stage.fit
    >>> Stage.safe('nonexistent')
    Stage.predict
    '''
    fit = auto()
    validate = auto()
    test = auto()
    predict = auto()
    
    @classmethod
    def _missing_(cls, val: str):
        return missing(cls, val, cls.predict)
    
    @classmethod
    def safe(cls, stage: 'Stage') -> 'Stage':
        '''Take a key and return the corresponding enum member if it exists, 
        otherwise return the default member'''
        try: return cls(stage)
        except ValueError: return cls.predict

# %% ../nbs/10_enum.ipynb 39
class Channels(StrEnum):
    '''
    Enumeration for different channel configurations in tensors, particularly for image data.

    Methods
    -------
    flip(t: Tensor) -> Tensor
        Flips the tensor between THWC and TCHW channel configurations.

    cidx(t: Tensor) -> int
        Returns the index of the channel dimension in the tensor.

    to(t: Tensor, c: 'Channels', cidx: int) -> Tensor
        Converts a tensor to the specified channel configuration.

    resize(t: Tensor, resize: ImgSizeQ, output_channels: 'Channels') -> Tensor
        Resizes the tensor and optionally changes its channel configuration.
    '''
    THWC = auto() # 0, 1, 2, 3
    TCHW = auto() # 0, 3, 1, 2
    
    @classmethod
    def flip(cls, t: tensor) -> tensor:
        '''Flips the tensor between THWC and TCHW channel configurations.'''
        return t.transpose(1, 3)
    
    @classmethod
    def cidx(cls, t: tensor) -> int:
        '''Returns the index of the channel dimension'''
        idx = -1
        for i in range(len(t.shape)):
            if t.shape[i] != 3: continue
            idx = i
        
        idx = idx if idx >= 0 else np.argmin(t.shape)
        if (val := t.shape[idx]) not in {1, 3}: 
            raise ValueError(f'Expected eitehr 1 or 3 channels, got {val}')
        return idx
    
    @classmethod
    def _missing_(cls, val: str): 
        return missing(cls, val, cls.THWC)
    
    @classmethod
    def safe(cls, c: 'Channels'):
        '''
        Safely retrieves the channel enumeration, defaulting to THWC if invalid.

        Parameters
        ----------
        c : Channels
            The channel configuration to validate.

        Returns
        -------
        Channels
            The validated channel configuration, defaults to THWC if input is invalid.
        '''
        try: return cls(c)
        except ValueError: return cls.THWC
        
    @classmethod
    def to(cls, t: tensor, c: 'Channels' = 'THWC', cidx: int = None) -> tensor:
        '''
        Converts a tensor to the specified channel configuration.

        Parameters
        ----------
        t : Tensor
            The input tensor to convert.
        c : Channels, optional
            The target channel configuration, defaults to THWC.
        cidx : int, optional
            The current channel index in the tensor, calculated if not provided.

        Returns
        -------
        Tensor
            The tensor converted to the specified channel configuration.
        '''
        c = cls.safe(c)
        if isnone(cidx): cidx = cls.cidx(t)
        if c == cls.THWC and cidx == 1: return cls.flip(t)
        if c == cls.THWC and cidx == 3: return t
        if c == cls.TCHW and cidx == 1: return t
        if c == cls.TCHW and cidx == 3: return cls.flip(t)
        return t
    
    @classmethod
    def resize(cls, t: tensor, resize: imagesizeq = None, output_channels: 'Channels' = 'THWC') -> tensor:
        '''
        Resizes a tensor and changes its channel configuration if needed.

        Parameters
        ----------
        t : Tensor
            The input tensor to resize.
        resize : ImgSizeQ, optional
            The target size for resizing, does nothing if None.
        output_channels : Channels, optional
            The channel configuration for the output tensor, defaults to THWC.

        Returns
        -------
        Tensor
            The resized tensor, potentially with a different channel configuration.
        '''
        if isnone(resize): return t 
        # F: ModuleType = imod('torchvision.transforms.functional')
        F: ModuleType = tv_functional.duck
        t = Channels.to(t, Channels.TCHW)
        t = F.resize(t, resize)
        t = Channels.to(t, output_channels)
        return t

# %% ../nbs/10_enum.ipynb 41
class NonLinearity(str, ModuleEnum, module = 'torch.nn', default='Tanh'):
    '''
    Enumeration of non-linear activation functions and other related layers in PyTorch.

    Methods
    -------
    get(*args, **kwargs) -> Layer
        Retrieves the corresponding PyTorch layer or activation function.
    '''
    # weighted sum, nonlinearity
    ELU = auto()
    Hardshrink = auto()
    Hardsigmoid = auto()
    Hardtanh = auto()
    Hardswish = auto()
    LeakyReLU = auto()
    LogSigmoid = auto()
    MultiheadAttention = auto()
    PReLU = auto()
    SELU = auto()
    CELU = auto()
    GELU = auto()
    Sigmoid = auto()
    SiLU = auto()
    Mish = auto()
    Softplus = auto()
    Softshrink = auto()
    Tanh = auto()
    Tanhshrink = auto()
    Threshold = auto()
    GLU = auto()
    
    # other
    Softmin = auto()
    Softmax = auto()
    Softmax2d = auto()
    LogSoftmax = auto()
    AdaptiveLogSoftmaxWithLoss = auto()
    
    Identity = auto()

# %% ../nbs/10_enum.ipynb 43
class InitMethod(str, ModuleEnum, module = 'torch.nn.init', default = 'xavier_normal'):
    '''
    Enumeration for different initialization methods for neural network layers in PyTorch.

    Methods
    -------
    get(*args, **kwargs) -> Layer
        Retrieves and applies the specified initialization method to a layer.
    '''
    constant = auto()
    dirac = auto()
    eye = auto()
    kaiming_normal = auto()
    normal = auto()
    orthogonal = auto()
    sparse = auto()
    uniform = auto()
    xavier_normal = auto()
    xavier_uniform = auto()
    zeros = auto()
     
    @classmethod
    def imp(cls: Type[T], key: T | None = None, *args: P.args, **kwargs: P.kwargs) -> Callable:
        '''Import the class or function corresponding to the enum member, 
        or the default member of the enum class'''
        mth = cls.safe(key)
        mod = cls.module()
        suffix = kwargs.pop(_SUFFIX, U1)
        return getattr(mod, f'{mth}{suffix}')

# %% ../nbs/10_enum.ipynb 45
class RecurrentLayer(str, ModuleEnum, module = 'torch.nn', default='LSTM'):
    '''
    Enumeration for different types of recurrent layers in PyTorch.

    Methods
    -------
    get(*args, **kwargs) -> Layer
        Retrieves the corresponding recurrent layer from PyTorch.
    '''
    GRU = auto()
    RNN = auto()
    LSTM = auto()

# %% ../nbs/10_enum.ipynb 47
DimChars = _DimChars
DimType  = _DimType
DimName  = _DimName
dimchars = _dimchars
size2str = _size2str

# %% ../nbs/10_enum.ipynb 49
class SliceFormat(StrEnum):
    '''
    Enumeration for representing different formats of slices used in tensor operations.

    This class provides methods to handle two common slice formats:
    - 'coord': where slices are represented as coordinates (e.g., (x0, y0, z0, ...), (x1, y1, z1, ...))
    - 'slice': where slices are represented as Python slice objects (e.g., (x0, x1), (y0, y1), (z0, z1), ...)

    Members
    -------
    coord : Auto-assigned Enum member
        Represents slices in coordinate format.
    slice : Auto-assigned Enum member
        Represents slices in Python slice object format.

    Methods
    -------
    flip(*slcs: Slcs) -> Slcs
        Flips the slice format from 'coord' to 'slice' or vice versa.

    safe(c: 'SliceFormat') -> 'SliceFormat'
        Safely returns a SliceFormat member, defaulting to 'slice' if input is invalid.

    guess(*slcs: Slcs) -> 'SliceFormat'
        Guesses the slice format based on the input slices.

    safeflip(*slcs: Slcs, fmt: 'SliceFormat') -> Slcs
        Safely flips the slice format, if necessary, to match the desired format.

    to(*slcs: Slcs, fmt: 'SliceFormat', cur: 'SliceFormat') -> Slcs
        Converts the given slices to the desired format, guessing the current format if not provided.

    Examples
    --------
    >>> SliceFormat.flip((0, 10), (0, 10))
    ((0, 0), (10, 10))

    >>> SliceFormat.safe('unknown_format')
    SliceFormat.slice

    >>> SliceFormat.to((0, 10), (0, 10), fmt=SliceFormat.coord)
    (0, 0), (10, 10)
    '''

    coord = auto()
    '''(x0, y0, z0, ...), (x1, y1, z1, ...), (x2, y2, z2, ...), ...'''
    slice = auto()
    '''(x0, x1, ), (y0, y1, ), (z0, z1, ), ...'''
    
    @classmethod
    def flip(cls, *slcs: slicespecs) -> slicespecs:
        '''
        Flips the format of the given slices between 'coord' and 'slice'.

        Parameters
        ----------
        slcs : Slcs
            The slices to flip. Can be in 'coord' or 'slice' format.

        Returns
        -------
        Slcs
            The slices flipped to the opposite format.

        Examples
        --------
        >>> SliceFormat.flip((0, 10), (0, 10))
        ((0, 0), (10, 10))
        '''
        return tuple(zip(*slcs))
    
    @classmethod
    def safe(cls, c: 'SliceFormat'):
        '''
        Safely returns a valid SliceFormat member.

        Parameters
        ----------
        c : SliceFormat
            The slice format to validate.

        Returns
        -------
        SliceFormat
            The validated slice format, defaults to 'slice' if input is invalid.

        Examples
        --------
        >>> SliceFormat.safe('unknown_format')
        SliceFormat.slice
        '''
        try: return cls(c)
        except ValueError: return cls.slice

    @classmethod
    def guess(cls, *slcs: slicespecs) -> slicespecs:
        '''
        Guesses the slice format based on the given slices.

        Parameters
        ----------
        slcs : Slcs
            The slices to evaluate.

        Returns
        -------
        SliceFormat
            The guessed slice format.

        Examples
        --------
        >>> SliceFormat.guess((0, 10), (0, 10))
        SliceFormat.slice
        '''
        if any(isslice(slc) for slc in slcs): return cls.slice
        if any([len(slc) > 3 for slc in slcs]): return cls.coord
        return cls.slice

    def safeflip(self, *slcs: slicespecs, fmt: 'SliceFormat') -> slicespecs:
        '''
        Safely flips the slice format to the specified format.

        Parameters
        ----------
        slcs : Slcs
            The slices to flip.
        fmt : SliceFormat
            The target format to flip to.

        Returns
        -------
        Slcs
            The slices in the specified format.

        Examples
        --------
        >>> SliceFormat.safeflip((0, 10), (0, 10), fmt=SliceFormat.coord)
        (0, 0), (10, 10)
        '''
        fmt = self.safe(fmt)
        if self == fmt: return slcs
        else: return self.flip(*slcs)

    @classmethod
    def to(cls, *slcs: slicespecs, fmt: 'SliceFormat', cur: 'SliceFormat' = None) -> slicespecs:
        '''
        Converts the given slices to the specified format.

        Parameters
        ----------
        slcs : Slcs
            The slices to convert.
        fmt : SliceFormat
            The target format for conversion.
        cur : SliceFormat, optional
            The current format of the slices, will be guessed if not provided.

        Returns
        -------
        Slcs
            The slices converted to the specified format.

        Examples
        --------
        >>> SliceFormat.to((0, 10), (0, 10), fmt=SliceFormat.coord)
        (0, 0), (10, 10)
        '''
        fmt = cls.safe(fmt)
        if isnone(cur): cur = cls.guess(*slcs)
        if cur == fmt: return slcs
        return cls.flip(*slcs)

# %% ../nbs/10_enum.ipynb 53
class DiffusionMethod(str, ModuleEnum, module = 'utrc.diff', default='diffusion'):
    '''Enumeration for specifying diffusion computation methods in differential geometry contexts.

    Attributes
    ----------
    diffusion : auto
        Uses diffusion process for geodesic computation.
        
    distance : auto
        Computes geodesic distances.
        
    diff_map : auto
        Computes diffusion maps.
        
    affinity : auto
        Computes diffusion affinities.
        
    phate : auto
        Uses PHATE (Potential of Heat-diffusion for Affinity-based Transition Embedding) for computation.

    Methods
    -------
    imp(cls, key=None, *args, **kwargs) -> Callable
        Dynamically imports and returns the computation function based on the geodesic method.

    Examples
    --------
    >>> Geodesic('diffusion')
    Geodesic.diffusion
    >>> Geodesic.imp('diffusion')  # Returns the corresponding function from `ptrc.diff`.
    '''
    Diffusion = auto()
    DiffusionMap = auto()
    DiffusionAffinity = auto()
    DiffusionDistance = auto()
    PHATEDistance = auto()
    
    @classmethod
    def _missing_(cls, val: str): 
        return missing(cls, val, cls.PHATEDistance)
    
    @classmethod
    def imp(
        cls: Type[T], 
        key: T | str | None = None, 
        __module: str | None = None,
        *args: P.args, **kwargs: P.kwargs
    ) -> Callable:
        geo = cls.safe(geo)
        mod = kwargs.get('__module', __module) or cls.module()
        fnc = getattr(mod, str(geo), None)
        return fnc

# %% ../nbs/10_enum.ipynb 55
class Geodesic(str, ModuleEnum, module = 'utrc.diff', default='diffusion'):
    '''Enumeration for specifying geodesic computation methods in differential geometry contexts.

    Attributes
    ----------
    diffusion : auto
        Uses diffusion process for geodesic computation.
        
    distance : auto
        Computes geodesic distances.
        
    diff_map : auto
        Computes diffusion maps.
        
    affinity : auto
        Computes diffusion affinities.
        
    phate : auto
        Uses PHATE (Potential of Heat-diffusion for Affinity-based Transition Embedding) for computation.

    Methods
    -------
    imp(cls, key=None, *args, **kwargs) -> Callable
        Dynamically imports and returns the computation function based on the geodesic method.

    Examples
    --------
    >>> Geodesic('diffusion')
    Geodesic.diffusion
    >>> Geodesic.imp('diffusion')  # Returns the corresponding function from `ptrc.diff`.
    '''
    diffusion = auto()
    distance = auto()
    diff_map = auto()
    affinity = auto()
    phate = auto()
    
        
    @classmethod
    def imp(
        cls: Type[T], 
        key: T | str | None = None, 
        __module: str | None = None,
        *args: P.args, **kwargs: P.kwargs
    ) -> Callable:
        geo = cls.safe(geo)
        mod = kwargs.get('__module', __module) or cls.module()
        match geo:
            case cls.diffusion: fnc = getattr(mod, 'Diffusion', None)                
            case cls.distance:  fnc = getattr(mod, 'DiffusionDistance', None)
            case cls.diff_map:  fnc = getattr(mod, 'DiffusionMap', None)
            case cls.affinity:  fnc = getattr(mod, 'DiffusionAffinity', None)
            case cls.phate:     fnc = getattr(mod, 'PHATEDistance', None)
            case _: fnc = getattr(mod, 'PHATEDistance', None)
        return fnc

# %% ../nbs/10_enum.ipynb 57
class Framework(StrEnum):
    '''Enumeration for specifying the tensor operation framework.

    Attributes
    ----------
    numpy : auto
        Indicates the use of NumPy for tensor operations.
        
    torch : auto
        Indicates the use of PyTorch for tensor operations.

    Methods
    -------
    _missing_(val: str) -> 'Framework'
        Provides a default framework for missing entries.
        
    safe(framework: 'Framework') -> 'Framework'
        Returns the corresponding enum member if it exists; otherwise, returns a default framework.
        
    swap(X, framework='numpy', device=None)
        Converts tensors between NumPy and PyTorch formats, optionally specifying a device for PyTorch tensors.

    Examples
    --------
    >>> Framework('numpy')
    Framework.numpy
    >>> Framework.swap(numpy_array, framework='torch')
    # Returns a PyTorch tensor converted from a NumPy array.
    '''
    numpy = auto()
    torch = auto()
    
    @classmethod
    def _missing_(cls, val: str) -> 'Framework':
        return missing(cls, val, cls.numpy)
    
    @classmethod
    def safe(cls, framework: 'Framework') -> 'Framework':
        try: return cls(framework)
        except ValueError: return cls.numpy
    
    @classmethod
    def swap(cls, X, framework: 'Framework' = NUMPY, device: deviceq = None):
        framework = cls.safe(framework)
        device = asdev(getattr(X, DEVICE, device))
        match framework:
            case cls.numpy:
                return asarr(X)
            
            case cls.torch:
                X = asten(X)
                if notnone(device): 
                    X = X.to(device)
                return X
            
            case _: ...
        return X

# %% ../nbs/10_enum.ipynb 59
class LocReturn(StrEnum):
    '''What kind of location to return after an operation.
    
    Attributes
    ----------
    loc : 'loc'
        The default index returned following an operation.
        
    bloc : 'bloc'
        A boolean index with the same number of rows as the calling object e.g.
        `pd.Series([True, True, False, ..., False, True])`
    
    iloc : 'iloc'
        An index of just the integer indexes for the rows of the calling object.
    
    bools : 'bloc'
        An alias for `bloc`.
    
    index : 'iloc'
        An alias for `iloc`.
    '''
    loc = auto()
    bloc = auto()
    iloc = auto()
    bools = bloc
    index = iloc

# %% ../nbs/10_enum.ipynb 60
class SampleCategoriesMethod(IntEnum):
    '''The order of operations for how to sample categorical data from a pandas DataFrame.
    
    Attributes
    ----------
    sample_split_stack : 1
        Groupby categories, then sample, split into the number of categories, then stack.
        This method is straightforward and effective for balanced data sets but may introduce overhead 
        for very large DataFrames due to the splitting and stacking operations.
    index_sample_stack : 2
        Get each category's indexes, subset, then sample, and finally stack.
        This method can be more efficient for large DataFrames with a significant number of categories, 
        as it operates directly on indexes and avoids the potential overhead of splitting and stacking.
    sample_group_stack : 3
        Groupby categories, then sample, get each group, then stack.
        This method is efficient for DataFrames where direct access to groups is faster and can be more memory efficient
        than the split and stack approach, especially for unevenly distributed categories.
    '''
    sample_split_stack = auto() # 1
    index_sample_stack = auto() # 2
    sample_group_stack = auto() # 3

# %% ../nbs/10_enum.ipynb 62
#| export



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/20_kwds.ipynb.

# %% auto 0
__all__ = ['diffusion_kwds', 'recurrent_kwds', 'recurrent_encoder_kwds', 'recurrent_decoder_kwds']

# %% ../nbs/20_kwds.ipynb 6
import inspect
from inspect import Signature, Parameter, get_annotations
from functools import wraps, partial

# %% ../nbs/20_kwds.ipynb 8
from typing import TypedDict, Type, Callable

# %% ../nbs/20_kwds.ipynb 11
#| export

# %% ../nbs/20_kwds.ipynb 13
#| export


# %% ../nbs/20_kwds.ipynb 15
try: import torch, torch.nn as nn
except ImportError: ...

# %% ../nbs/20_kwds.ipynb 17
#| export


# %% ../nbs/20_kwds.ipynb 19
from atyp import DTypeQ, DeviceQ
from dtyp import setkeywords

# %% ../nbs/20_kwds.ipynb 21
from .atyp import (RecurrentAutoEncoderKeywords, DiffusionKeywords)
from .enum import NonLinearity, RecurrentLayer

# %% ../nbs/20_kwds.ipynb 25
def diffusion_kwds(
    knn: int = 5,
    tmax: int = 5,
    anisotropy: int = 1,
    tdiff: int = 1,
    topeig: int = 100,
    log: bool = False,
    normalize: bool = False,
    symmetrize: bool = False,
    verbose: bool = False,
    nemb: int = 10,
    njobs: int = -1,
    device: DeviceQ = None,
) -> DiffusionKeywords:
    params = dict(
        knn=knn, tmax=tmax, anisotropy=anisotropy, tdiff=tdiff, topeig=topeig, log=log, 
        normalize=normalize, symmetrize=symmetrize, verbose=verbose, nemb=nemb, njobs=njobs, 
        device=device,
    )
    return params

# %% ../nbs/20_kwds.ipynb 27
def recurrent_kwds(
    input_size: int, hidden_size: int, num_layers: int = 1, 
    bias: bool = True, batch_first: bool = True, dropout: float = 0.2, 
    bidirectional: bool = False, proj_size: int = 0, 
    nonlinearity: NonLinearity = NonLinearity.Tanh, 
    device: DeviceQ = None,  dtype: DTypeQ = None, 
    kind: RecurrentLayer = RecurrentLayer.LSTM,
):
    '''
    Generate keyword arguments for initializing a recurrent neural network layer.

    Parameters
    ----------
    input_size, hidden_size, num_layers, bias, batch_first, dropout, bidirectional, proj_size, nonlinearity, device, dtype, kind : 
        Parameters for defining the RNN layer properties.

    Returns
    -------
    dict
        A dictionary of keyword arguments for RNN layer initialization.

    Examples
    --------
    >>> recurrent_kwds(10, 20, num_layers=2, bidirectional=True)
    # Returns keyword arguments for a bidirectional RNN with 2 layers.
    '''
    params = dict(
        input_size=input_size, hidden_size=hidden_size, num_layers=num_layers, bias=bias, 
        batch_first=batch_first, dropout=dropout, bidirectional=bidirectional,
        proj_size=proj_size, nonlinearity=nonlinearity, device=device, dtype=dtype,
    )
    
    if RecurrentLayer(kind) != RecurrentLayer.RNN:  params.pop('nonlinearity', None)
    if RecurrentLayer(kind) != RecurrentLayer.LSTM: params.pop('proj_size', None)

    return params

# %% ../nbs/20_kwds.ipynb 30
def recurrent_encoder_kwds(**kwargs: RecurrentAutoEncoderKeywords):
    '''
    Generate keyword arguments for initializing an encoder in a recurrent autoencoder.

    Parameters
    ----------
    kwargs : RecurrentAutoEncoderKeywords
        Keyword arguments specific to the encoder.

    Returns
    -------
    dict
        A dictionary of keyword arguments for encoder initialization.

    Examples
    --------
    >>> recurrent_encoder_kwds(input_size=10, hidden_size=20)
    # Returns keyword arguments for an encoder.
    '''
    result, params = dict(), kwargs.copy()
    input_size  = params.pop('input_size',  None)
    hidden_size = params.pop('hidden_size', None)
    output_size = params.pop('output_size', None)
    result.setdefault('input_size',  input_size or hidden_size)
    result.setdefault('hidden_size', output_size or hidden_size)
    return {**result, **params}

# %% ../nbs/20_kwds.ipynb 32
def recurrent_decoder_kwds(**kwargs: RecurrentAutoEncoderKeywords):
    '''
    Generate keyword arguments for initializing a decoder in a recurrent autoencoder.

    Parameters
    ----------
    kwargs : RecurrentAutoEncoderKeywords
        Keyword arguments specific to the decoder.

    Returns
    -------
    dict
        A dictionary of keyword arguments for decoder initialization.

    Examples
    --------
    >>> recurrent_decoder_kwds(input_size=10, hidden_size=20)
    # Returns keyword arguments for a decoder.
    '''
    result, params = dict(), kwargs.copy()
    input_size  = params.pop('input_size',  None)
    hidden_size = params.pop('hidden_size', None)
    output_size = params.pop('output_size', None)
    result.setdefault('hidden_size', input_size  or hidden_size)
    result.setdefault('output_size', output_size or hidden_size)
    return {**result, **params}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_atyp.ipynb.

# %% auto 0
__all__ = ['T', 'P', 'Mixin', 'MetaClass']

# %% ../nbs/01_atyp.ipynb 6
from abc import ABCMeta
from enum import EnumMeta

# %% ../nbs/01_atyp.ipynb 8
from typing import (Any, Type, Union, Tuple, TypeAlias, Callable, ParamSpec, TypeVar)

# %% ../nbs/01_atyp.ipynb 10
#| export


# %% ../nbs/01_atyp.ipynb 12
#| export


# %% ../nbs/01_atyp.ipynb 14
#| export


# %% ../nbs/01_atyp.ipynb 16
T = TypeVar('T')
P = ParamSpec('P')

Mixin: TypeAlias = Tuple[Type, Callable[P, Any]]

MetaClass: TypeAlias = Type[Union[T, EnumMeta, ABCMeta]]

# %% ../nbs/01_atyp.ipynb 18
#| export



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_cons.ipynb.

# %% auto 0
__all__ = ['TORCH_NN']

# %% ../nbs/00_cons.ipynb 6
#| export


# %% ../nbs/00_cons.ipynb 8
#| export


# %% ../nbs/00_cons.ipynb 10
#| export


# %% ../nbs/00_cons.ipynb 12
from nlit import SUFFIX, MODULE, DEFAULT, _MEMBER_NAMES_, _MEMBER_MAP_
from uscr import private

# %% ../nbs/00_cons.ipynb 14
#| export


# %% ../nbs/00_cons.ipynb 16
TORCH_NN = 'torch.nn'

# %% ../nbs/00_cons.ipynb 17
_MODULE = private(MODULE)
_DEFAULT = private(DEFAULT)
_SUFFIX = private(SUFFIX)

# %% ../nbs/00_cons.ipynb 19
#| export



__version__ = "0.0.17"
# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00__init__.ipynb.

# %% auto 0
__all__ = ['loadmod']

# %% ../nbs/00__init__.ipynb 3
from itertools import chain
from types import ModuleType
from importlib import import_module

def loadmod(mod: ModuleType):
    if hasattr(mod, '__all__'): return mod.__all__
    if hasattr(mod, '__module__'): return loadmod(import_module(mod.__module__))
    raise AttributeError(f"module {mod} has no attribute '__all__'")

# %% ../nbs/00__init__.ipynb 4
from . import (cons, atyp, util, meta, emod, solv, enum,)
from .cons import *
from .atyp import *
from .util import *
from .meta import *
from .emod import *
from .solv import *
from .enum import *

# %% ../nbs/00__init__.ipynb 5
__all__ = sorted(set(chain(*[
    loadmod(mod) for mod in (cons, atyp, util, meta, emod, solv, enum,)
])))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_meta.ipynb.

# %% auto 0
__all__ = ['ModuleEnumMeta']

# %% ../nbs/03_meta.ipynb 6
import copy
from enum import EnumMeta, _EnumDict, member

# %% ../nbs/03_meta.ipynb 8
from typing import (Type, TypeGuard)

# %% ../nbs/03_meta.ipynb 10
#| export


# %% ../nbs/03_meta.ipynb 12
from nlit import MIXINS, ASMEMBER, MODULE, DEFAULT, _MEMBER_NAMES_, _MEMBER_MAP_

# %% ../nbs/03_meta.ipynb 14
from .cons import TORCH_NN, _MODULE, _DEFAULT
from .atyp import T, P, Mixin
from .util import enumdict, evalue, missing

# %% ../nbs/03_meta.ipynb 16
class ModuleEnumMeta(EnumMeta):
    def _check_for_existing_members_(class_name, bases):
        '''Check for existing members in base classes.'''
        
    @classmethod
    def duplicate(mcls: Type[T], v: T, *args: P.args, **kwargs: P.kwargs):
        '''Duplicate a value within the enumeration dictionary.'''
        mixins: list[Mixin] = kwargs.get(MIXINS, getattr(mcls, MIXINS, []))
        asmemb = kwargs.get(ASMEMBER, False)
        copied = copy.deepcopy(evalue(v))
        for (mixtype, hook) in mixins:
            if not isinstance(copied, mixtype): continue
            copied = hook(copied, *args, **kwargs)
        return member(copied) if asmemb else copied
        
    @classmethod
    def __copydict__(mcls: Type[T], clsdict: _EnumDict, *args: P.args, **kwargs: P.kwargs) -> _EnumDict:
        '''Create a deep copy of the enumeration dictionary.'''
        newdict = enumdict(clsdict._cls_name)
        for k, v in clsdict.items(): 
            newdict[k] = mcls.duplicate(v, *args, **kwargs)
        return newdict
    
    @classmethod
    def __prepare__(mcls: Type[T], name: str, bases: tuple, **kwargs) -> _EnumDict:
        '''Prepare the enumeration dictionary before class creation.'''
        clsdict = enumdict(name)
        return clsdict
    
    def __new__(mcls: Type[T], name, bases, clsdict, **kwargs):
        '''Create a new enumeration class.'''
        newcls = super().__new__(mcls, name, bases, clsdict)
        setattr(newcls, _MODULE, kwargs.get(MODULE, None))
        setattr(newcls, _DEFAULT, kwargs.get(DEFAULT, None))
        return newcls

    def _default_(cls: Type[T]) -> T:
        '''Get the default value for the enumeration class.'''
        keys = getattr(cls, _MEMBER_NAMES_, [None])
        return getattr(cls, _DEFAULT, keys[0])
    
    def _module_(cls: Type[T]) -> str:
        '''Get the module name for the enumeration class.'''
        return getattr(cls, _MODULE, TORCH_NN)
    
    @classmethod
    def _missing_(cls: Type[T], val: str | T, default = None) -> T:
        '''Method called for missing members during lookup.

        Parameters
        ----------
        cls : Enum
            The enumeration class.
        
        val : str
            The value to check for in the enum members.
        

        Returns
        -------
        Enum member or default
            The matched enum member or the default value.
        '''
        if default is None: default = cls._default_()
        return missing(cls, val, default)
    
    def __contains__(self: T, key: str) -> TypeGuard[T]:
        '''Check if the key exists in the enumeration's members.'''
        
        attrs = (_MEMBER_MAP_, _MEMBER_NAMES_)
        for (attr, case) in zip(attrs, (True, False)):
            for mem in getattr(self, attr, []):
                if case and (str(mem).casefold() == key.casefold()) or (mem == key):
                    return True
        return False
    
    def __call__(self: T, key: str | None = None, default = None) -> T:
        '''Get the enum member for the given key.'''
        if key is None: key = self._default_()
        if key not in self: key = self._missing_(key, default)
        return self[key]

# %% ../nbs/03_meta.ipynb 18
#| export



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_emod.ipynb.

# %% auto 0
__all__ = ['ModuleEnum']

# %% ../nbs/04_emod.ipynb 6
from enum import Enum
from importlib import import_module

# %% ../nbs/04_emod.ipynb 8
from types import ModuleType
from typing import (Type, Callable)

# %% ../nbs/04_emod.ipynb 10
#| export


# %% ../nbs/04_emod.ipynb 12
#| export

# %% ../nbs/04_emod.ipynb 14
from .atyp import T, P
from .meta import ModuleEnumMeta
from .util import missing

# %% ../nbs/04_emod.ipynb 16
class ModuleEnum(Enum, metaclass=ModuleEnumMeta):
    @classmethod
    def _missing_(cls: Type[T], val: str | T, default = None) -> T:
        '''Method called for missing members during lookup.'''
        if default is None: default = cls._default_()
        return missing(cls, val, default)
    
    @classmethod
    def safe(cls: Type[T], key: T) -> T:
        '''Take a key and return the corresponding enum member if it exists, otherwise return the default member'''
        try: return cls(key)
        except ValueError: ...
        try: return cls(cls._default_())
        except ValueError: cls._default_()
        
    @classmethod
    def module(cls: Type[T]) -> ModuleType:
        '''Return the module from which to load members corresponding to the enum class'''
        try: return import_module(cls._module_())
        except: return None
         
    @classmethod
    def imp(cls: Type[T], key: T | None = None, *args: P.args, **kwargs: P.kwargs) -> Callable:
        '''Import the class or function corresponding to the enum member, or the default member of the enum class'''
        ins = cls.safe(key)
        mod = cls.module()
        if isinstance(ins, cls): ins = ins.name
        if not isinstance(ins, str): ins = str(ins)
        return getattr(mod, ins)
    
    @classmethod
    def init(cls: Type[T], key: T | None = None, *args: P.args, **kwargs: P.kwargs):
        '''Initalize the class or function corresponding to the enum member, or the default member of the enum class with the given arguments'''
        fn = cls.imp(key, *args, **kwargs)
        return fn(*args, **kwargs)
    
    def get(self: T, *args: P.args, **kwargs: P.kwargs) -> Callable:
        '''Import the class or function of this enum member'''
        return self.imp(self, *args, **kwargs)
    
    def make(self: T, *args: P.args, **kwargs: P.kwargs):
        '''Initalize the class or function corresponding with the given arguments of this enum member'''
        return self.init(self, *args, **kwargs)

# %% ../nbs/04_emod.ipynb 18
#| export



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_util.ipynb.

# %% auto 0
__all__ = ['nextv', 'enumdict', 'evalue', 'missing', 'get_torch_module']

# %% ../nbs/02_util.ipynb 6
from enum import Enum, _EnumDict
from importlib import import_module

# %% ../nbs/02_util.ipynb 8
from types import ModuleType
from typing import (Type)

# %% ../nbs/02_util.ipynb 10
#| export


# %% ../nbs/02_util.ipynb 12
from nchr import NIL
from nlit import NAME, VALUE, _DEFAULT_
from quac import layer, boolq
from chck import isnone, notnone

# %% ../nbs/02_util.ipynb 14
from .cons import TORCH_NN
from .atyp import T

# %% ../nbs/02_util.ipynb 16
def nextv(name: str, start: int, count: int, last_values: list[str]):
    '''get the next value for an enum
    
    Parameters
    ----------
    name : str
        the name of the enum member
    start : int
        the starting value for the enum
        
    count : int
        the number of members in the enum
        
    last_values : list[str]
        the last values of the enum members
    
    Returns
    -------
    str
        the next value for the enum member
    
    See Also
    --------
    _EnumDict._generate_next_value : the default implementation
    '''
    return name


def enumdict(name: str) -> _EnumDict:
    '''create a new enum dict
    
    Parameters
    ----------
    name : str
        the name of the enum class
        
    Returns
    -------
    _EnumDict
        the new enum dict
        
    Notes
    -----
    This function is a workaround for the fact that the `Enum` class does not expose the `_EnumDict` class.
    Thus the paramter `name` is used to set the `_cls_name` attribute of the new enum dict, which is required.
    
    '''
    newdict = _EnumDict()
    newdict._cls_name = name
    newdict._generate_next_value = nextv
    # try: newdict._generate_next_value = StrEnum._generate_next_value_
    # except: newdict._generate_next_value = Enum._generate_next_value_
    return newdict

def evalue(e: Enum): 
    '''Get the value of an enum member or return the enum member itself.
    This is useful if you are unsure if `e` is the enum or an enum member.
    '''
    return getattr(e, VALUE, e)

# %% ../nbs/02_util.ipynb 17
def missing(enm: Type[T], val: str | T, default = None) -> T:
    '''Method called for missing members during lookup.

    Parameters
    ----------
    cls : Enum
        The enumeration class.
    
    val : str
        The value to check for in the enum members.
    

    Returns
    -------
    Enum member or default
        The matched enum member or the default value.
    '''
    # for mem in enm:
    #     mal: member = getattr(mem, VALUE, '')
    #     if str(mal).casefold() == str(val).casefold(): return mem
    # if hasattr(enm, '_default_'): return enm._default_() or default
    # return default

    for mem in enm:
        mal = getattr(mem, VALUE, NIL)        
        if str(mal).casefold() == str(val).casefold(): return mem
    if hasattr(enm, _DEFAULT_): 
        edef = enm._default_()
        try: return enm(edef)
        except: ...
        try: return enm(default)
        except: return default
    return default

# %% ../nbs/02_util.ipynb 18
def get_torch_module(
    obj, *args, __module: ModuleType = TORCH_NN, 
    __enumkey: boolq = None, __enumval: boolq = None, __suffix: str = NIL,
    **kwargs
) -> layer:
    '''
    Fetch a torch module based on an enumeration value or its name.

    Parameters
    ----------
    obj : Enum
        An enumeration instance or class.
    args : tuple
        Positional arguments to pass to the module constructor.
    __module : ModuleType, optional
        The torch module to fetch from, default is `torch.nn`.
        
    __enumkey : BoolQ, optional
        If True, use the enum's name for fetching, otherwise use the value.
        
    __enumval : BoolQ, optional
        If True, use the enum's value for fetching, otherwise use the name.
        
    __suffix : str, optional
        Suffix to append to the enum value or name.
        
    kwargs : dict
        Keyword arguments to pass to the module constructor.

    Returns
    -------
    Layer
        The instantiated torch module.
    '''
    # ensure that only one of __usekey and __useval is True
    usekey, useval = None, None
    if notnone(__enumkey): usekey = __enumkey
    if notnone(__enumval): useval = __enumval
    if isnone(usekey) and isnone(useval): usekey = True
    
    usekey = not useval if notnone(useval) else usekey
    useval = not usekey if notnone(usekey) else useval
    
    if issubclass(type(obj), Enum):
        val = getattr(obj, NAME if usekey else VALUE, obj.value)
    else: val = obj
        
    val = f'{val}{__suffix}'
    
    if __module is None: __module = TORCH_NN
    if isinstance(__module, str): __module = import_module(__module)
    mod = getattr(__module, val)
    if args or kwargs: mod = mod(*args, **kwargs)
    return mod

# %% ../nbs/02_util.ipynb 20
#| export
